"""
Utility functions for the compression
"""

import os
import gc
import pathlib
import numpy as np
import pandas as pd
import h5py
import scanpy as sc

from scquill.utils.types import _infer_dtype


def write_to_h5(
    fn_out,
    compressed_atlas,
    compression=22,
    quantisation="chromatin_accessibility",
    # chunked=True,
):
    """Store compressed atlas into h5 file.

    Args:
        fn_out: The h5 file with the compressed atlas.
        compressed_atlas: The dict with the result.
        groupby_order: The order of groups, typically cell types with potentially additionaal
            metadata such as time point and disease, that constitute the biological unit of
            approximation.
        quantisation: If not None, average measurement is quantised with these bins.
        compression: Use zstd compression of the data arrays (avg and frac). Levels are 1-22,
            whereas 0 or False means no compression. No performace decrease is observed.
    """
    add_kwargs = {}

    # Optional zstd compression using hdf5plugin
    if compression:
        try:
            import hdf5plugin

            # NOTE: decompressing zstd is equally fast no matter how much compression.
            # As for compression speed, levels 1-19 are normal, 20-22 "ultra".
            # A quick runtime test shows *faster* access for clevel=22 than clevel=3,
            # while the file size is around 10% smaller. Compression speed is significantly
            # slower, but (i) still somewhat faster than actually averaging the data and
            # (ii) compresses whole human RNA+ATAC is less than 1 minute. That's nothing
            # considering these approximations do not change that often.
            comp_kwargs = hdf5plugin.Zstd(clevel=compression)
        except ImportError:
            # TODO: throw some warning
            compression = False
            comp_kwargs = {}
    else:
        comp_kwargs = {}

    with h5py.File(fn_out, "a") as h5_data:
        # Version
        import scquill

        h5_data.attrs["scquill_version"] = scquill.__version__

        # Umbrella group for measurments (not metadata)
        me_all = h5_data.create_group("measurements")

        for measurement_type, approximation_dict_mt in compressed_atlas.items():

            # TODO: move the quantisation bit somewhere else
            # Data can be quantised for further compression (typically ATAC-Seq)
            if (quantisation == True) or (quantisation == measurement_type):
                if measurement_type == "chromatin_accessibility":
                    # NOTE: tried using quantiles for this, but they are really messy
                    # and subject to aliasing effects. 8 bits are more than enough for
                    # most biological questions given the noise in the data
                    qbits = 8
                    bins = np.array(
                        [-0.001, 1e-8]
                        + np.logspace(-4, 0, 2**qbits - 1).tolist()[:-1]
                        + [1.1]
                    )
                    # bin "centers", aka data quantisation
                elif measurement_type == "gene_expression":
                    # Counts per ten thousand quantisation
                    qbits = 16
                    bins = np.array(
                        [-0.001, 1e-8]
                        + np.logspace(-2, 4, 2**qbits - 1).tolist()[:-1]
                        + [1.1e4]
                    )
                else:
                    raise ValueError(f"Quantisation for {measurement_type} not set.")

                quantisation_array = (
                    [0] + np.sqrt(bins[1:-2] * bins[2:-1]).tolist() + [1]
                )

                qbytes = qbits // 8
                # Add a byte if the quantisation is not optimal
                if qbits not in (8, 16, 32, 64):
                    qbytes += 1
                avg_dtype = f"u{qbytes}"
                quantisation = True
            else:
                avg_dtype = "f4"
                quantisation = False

            features = approximation_dict_mt["var_names"].tolist()

            # Subgroup for a specific measurement type (multi-omics will have multiple)
            me = me_all.create_group(measurement_type)

            me.attrs["compression"] = compression

            me.create_dataset("var_names", data=np.array(features).astype("S"))
            if quantisation:
                me.create_dataset(
                    "quantisation", data=np.array(quantisation_array).astype("f4")
                )

            # Number of cells
            ncells = approximation_dict_mt["obs"]["cell_count"]
            me.create_dataset("cell_count", data=ncells.values, dtype="i8")

            # Grouping information and metadata
            obs = approximation_dict_mt["obs"]
            groupby_columns = obs.columns[:-1]
            groupby = me.create_group("groupby")
            groupby.attrs["n_levels"] = len(groupby_columns)
            groupby_names = groupby.create_group("names")
            groupby_types = groupby.create_group("dtypes")
            for i, col in enumerate(groupby_columns):
                groupby_names.attrs[str(i)] = col
                groupby_types.attrs[str(i)] = str(obs[col].dtype)
            meta = me.create_group("obs")
            for i, column in enumerate(obs.columns):
                dtype_store = _infer_dtype(obs[column].dtype)
                meta.create_dataset(
                    column,
                    data=obs[column].values.astype(dtype_store),
                )

            # Average in a cell type
            avg = approximation_dict_mt["Xave"]
            if quantisation:
                # pd.cut wants one dimensional arrays so we ravel -> cut -> reshape
                avg_vals = (
                    pd.cut(avg.ravel(), bins=bins, labels=False)
                    .reshape(avg.shape)
                    .astype(avg_dtype)
                )

            # TODO: manual chunking might increase performance a bit, the data is
            # typically accessed only vertically (each feature its own island)
            # if chunked:
            #    # Chunk each feature on its own: this is perfect for ATAC-Seq
            #    add_kwargs['chunks'] = (1, len(features))

            # Groups (cell types w/ or w/o additional metadata)
            obs_names = approximation_dict_mt["obs_names"]
            me.create_dataset(
                "obs_names",
                data=obs_names.values.astype("S"),
            )

            # Average and fraction detected
            me.create_dataset(
                "average",
                data=avg.T,
                dtype=avg_dtype,
                **add_kwargs,
                **comp_kwargs,
            )
            if measurement_type == "gene_expression":
                # Fraction detected in a cell type
                frac = approximation_dict_mt["Xfrac"]
                me.create_dataset(
                    "fraction",
                    data=frac.T,
                    dtype="f4",
                    **add_kwargs,
                    **comp_kwargs,
                )

            # Local neighborhoods, if available
            if "neighborhood" in approximation_dict_mt:
                neid = approximation_dict_mt["neighborhood"]
                neigroup = me.create_group("neighborhood")
                ncells = neid["cell_count"]
                neigroup.create_dataset("cell_count", data=ncells, dtype="i8")

                avg = neid["Xave"]
                if quantisation:
                    # pd.cut wants one dimensional arrays so we ravel -> cut -> reshape
                    avg_vals = (
                        pd.cut(avg.ravel(), bins=bins, labels=False)
                        .reshape(avg.shape)
                        .astype(avg_dtype)
                    )
                # NOTE: these are just "neighborhood 0" etc. so no need for fancy joining
                neigroup.create_dataset("obs_names", data=neid["obs_names"].astype("S"))
                neigroup.create_dataset(
                    "average",
                    data=avg.T,
                    dtype=avg_dtype,
                    **add_kwargs,
                    **comp_kwargs,
                )
                if measurement_type == "gene_expression":
                    # Fraction detected in a cell type
                    frac = neid["Xfrac"]
                    neigroup.create_dataset(
                        "fraction",
                        data=frac.T,
                        dtype="f4",
                        **add_kwargs,
                        **comp_kwargs,
                    )

                # Centroid coordinates
                coords_centroids = neid["coords_centroid"]
                neigroup.create_dataset(
                    "coords_centroid",
                    data=coords_centroids.T,
                    dtype=avg_dtype,
                    **add_kwargs,
                    **comp_kwargs,
                )

                # Convex hulls
                convex_hulls = neid["convex_hull"]
                hullgroup = neigroup.create_group("convex_hull")
                for ih, hull in enumerate(convex_hulls):
                    hullgroup.create_dataset(
                        str(ih),
                        data=hull,
                        dtype="f4",
                        **add_kwargs,
                        **comp_kwargs,
                    )



"""
Utility functions for the compression
"""

import gc
import numpy as np
import pandas as pd
import scanpy as sc


from .biological import _compress_biological
from .neighborhood import _compress_neighborhood


def approximate_dataset(
    adata,
    celltype_column,
    additional_groupby_columns=tuple(),
    measurement_type="gene_expression",
    include_neighborhood=True,
):
    """Compress atlas for one tissue after data is clean, normalised, and reannotated."""

    # Celltype averages
    biod = _compress_biological(
        adata,
        celltype_column,
        additional_groupby_columns,
        measurement_type,
    )
    features = adata.var_names
    result = {
        "var_names": features,
        "obs": biod["obs"],
        "obs_names": biod["obs_names"],
        "Xave": biod["Xave"],
    }
    if measurement_type == "gene_expression":
        result["Xfrac"] = biod["Xfrac"]

    if include_neighborhood:
        # Local neighborhoods
        neid = _compress_neighborhood(
            adata,
            biod["obs"],
            celltype_column,
            additional_groupby_columns,
            measurement_type,
        )
        result["neighborhood"] = neid

    compressed_atlas = {
        measurement_type: result,
    }

    return compressed_atlas



"""Read from files."""

import h5py
import hdf5plugin
import numpy as np
import pandas as pd
import anndata

from scquill.utils.types import _infer_dtype


def _order_reorder_features(dataset, var_idx, features, axis):
    """Efficiently extract unordered slices from an HDF5 dataset."""
    if var_idx is None:
        return dataset[:]

    idx1 = var_idx["ordered"].values
    idx2 = var_idx.loc[features, "reordered"].values
    if axis == 0:
        return dataset[idx1][idx2]
    else:
        return dataset[:, idx1][:, idx2]


def read_h5_to_anndata(
    h5_data,
    neighborhood,
    measurement_type,
    features=None,
):
    """Get an AnnData object in which each observation is an average."""
    if measurement_type not in h5_data["measurements"]:
        raise KeyError("Measurement type not found: {measurement_type}")

    me = h5_data["measurements"][measurement_type]
    compression = me.attrs.get("compression", None)

    if compression:
        try:
            import hdf5plugin
        except ImportError:
            raise ImportError(
                'You need the "hdf5plugin" package to decompress this approximation. You can install it e.g. via pip install hdf5plugin.'
            )

    var_names = me["var_names"].asstr()[:]
    if features is not None:
        var_idx = (
            pd.Series(np.arange(len(var_names)), index=var_names)
            .loc[features]
            .sort_values()
            .to_frame("ordered")
        )
        var_idx["reordered"] = np.arange(len(var_idx))
        var_names = features
    else:
        var_idx = None

    # NOTE: there are a few older versions around
    if "groupby" in me:
        groupby_names = []
        groupby_dtypes = []
        n_levels = me["groupby"].attrs["n_levels"]
        for i in range(n_levels):
            groupby_names.append(me["groupby"]["names"].attrs[str(i)])
            groupby_dtypes.append(me["groupby"]["dtypes"].attrs[str(i)])
        groupby = "\t".join(groupby_names)
    else:
        gbykey = "grouped_by"
        keys = list(me[gbykey].keys())
        if len(keys) != 1:
            raise ValueError("Expected exactly one groupby key, found: {keys}")
        groupby = keys[0]
        groupby_names = me[gbykey][groupby]["names"].asstr()[:]
        groupby_dtypes = me[gbykey][groupby]["dtypes"].asstr()[:]
        n_levels = len(groupby_names)

    if neighborhood:
        # TODO: read from neighborhood for flat groups (e.g. atlasapprox)
        neigroup = me["neighborhood"]
        Xave = _order_reorder_features(neigroup["average"], var_idx, features, axis=1)
        if "quantisation" in me:
            quantisation = me["quantisation"][:]
            Xave = quantisation[Xave]

        groupby_order = me["obs_names"].asstr()[:]
        obs_names = neigroup["obs_names"].asstr()[:]
        ncells = neigroup["cell_count"][:]
        coords_centroid = neigroup["coords_centroid"][:]
        convex_hulls = []
        for ih in range(len(coords_centroid)):
            convex_hulls.append(neigroup["convex_hull"][str(ih)][:])

        if measurement_type == "gene_expression":
            Xfrac = _order_reorder_features(
                neigroup["fraction"], var_idx, features, axis=1
            )
            adata = anndata.AnnData(
                X=Xave,
                layers={
                    "average": Xave,
                    "fraction": Xfrac,
                },
            )
        else:
            adata = anndata.AnnData(X=Xave)

        adata.obsm["X_ncells"] = ncells
        adata.obsm["X_umap"] = coords_centroid
        adata.uns["convex_hulls"] = convex_hulls

        adata.obs_names = pd.Index(obs_names, name="neighborhoods")
        adata.var_names = pd.Index(var_names, name="features")

    else:
        # NOTE: this is the historical way to store atlas approximations
        # when the organism has multiple tissues
        if "->" in groupby:
            read_fun = _read_data_from_stratified_group
        else:
            read_fun = _read_data_from_flat_group
        resdict = read_fun(
            me,
            groupby,
            groupby_names,
            groupby_dtypes,
            measurement_type,
            var_idx,
            features,
        )

        if measurement_type == "gene_expression":
            adata = anndata.AnnData(
                X=resdict["Xave"],
                obs=resdict["obs"],
                layers={
                    "average": resdict["Xave"],
                    "fraction": resdict["Xfrac"],
                },
            )
        else:
            adata = anndata.AnnData(
                X=resdict["Xave"],
                obs=resdict["obs"],
            )

        adata.var_names = pd.Index(var_names, name="features")

    adata.uns["approximation_groupby"] = {
        "names": groupby_names,
        "dtypes": groupby_dtypes,
    }
    if neighborhood:
        adata.uns["approximation_groupby"]["order"] = groupby_order
        adata.uns["approximation_groupby"]["cell_count"] = me["cell_count"][:]

    return adata


def _read_data_from_flat_group(
    me,
    groupby,
    groupby_names,
    groupby_dtypes,
    measurement_type,
    var_idx,
    features,
):
    # Data
    Xave = _order_reorder_features(me["average"], var_idx, features, axis=1)
    if "quantisation" in me:
        quantisation = me["quantisation"][:]
        Xave = quantisation[Xave]
    if measurement_type == "gene_expression":
        Xfrac = _order_reorder_features(me["fraction"], var_idx, features, axis=1)

    # Obs metadata
    obs_names = me["obs_names"].asstr()[:]
    obs = pd.DataFrame([], index=obs_names)
    for column, dtype in zip(groupby_names, groupby_dtypes):
        if _infer_dtype(dtype) == "S":
            obs[column] = me["obs"][column].asstr()[:]
        else:
            obs[column] = me["obs"][column][:]
    obs["cell_count"] = me["cell_count"][:]
    obs.index = pd.Index(obs.index, name=groupby)

    resdict = {
        "Xave": Xave,
        "obs": obs,
    }
    if measurement_type == "gene_expression":
        resdict["Xfrac"] = Xfrac
    return resdict


def _read_data_from_stratified_group(
    me,
    groupby,
    groupby_names,
    groupby_dtypes,
    measurement_type,
    var_idx,
    features,
):
    """Read data from a stratified group.

    NOTE: This is the historical way to store atlas approximations. It groups data by tissue (to have distinct neighborhoods)
    and then the obs_names inside each subgroup are the cell types. This function is therefore quite ad-hoc but it's ok for now.
    """
    # Iterate over the higher-level groups (tissues)
    Xave = []
    if measurement_type == "gene_expression":
        Xfrac = []
    obs = []

    # E.g. "tissues"
    groupby = "->".join(groupby_names)
    group = me["data"][groupby]
    for subgroupname, subgroup in group.items():
        # Data
        subXave = _order_reorder_features(
            subgroup["average"], var_idx, features, axis=1
        )
        if "quantisation" in me:
            quantisation = me["quantisation"][:]
            subXave = quantisation[subXave]
        Xave.append(subXave)
        if measurement_type == "gene_expression":
            Xfrac.append(
                _order_reorder_features(subgroup["fraction"], var_idx, features, axis=1)
            )

        # Obs metadata (notice the tricky obs_names indexing)
        subobs_names = subgroup["obs_names"].asstr()[:]
        subobs = pd.DataFrame([], index=subobs_names)
        # Set the common column (typically tissue)
        subobs[groupby_names[0]] = subgroupname
        # Set the discriminative column (typicall cell type)
        subobs[groupby_names[1]] = subobs_names
        # NOTE: assuming there are not other columns, i.e. something like tissue->celltype\tage is not supported
        subobs["cell_count"] = subgroup["cell_count"][:]
        subobs.index = subobs[groupby_names].apply("->".join, axis=1)
        obs.append(subobs)

    # Concatenate the lists
    Xave = np.concatenate(Xave)
    obs = pd.concat(obs)
    obs.index = pd.Index(obs.index, name=groupby)
    resdict = {
        "Xave": Xave,
        "obs": obs,
    }

    if measurement_type == "gene_expression":
        Xfrac = np.concatenate(Xfrac)
        resdict["Xfrac"] = Xfrac

    return resdict



import numpy as np
import anndata


def coarse_grain_anndata(adata, groupby):
    """Coarse grain an approximation further to broader grouping."""
    if "approximation_groupby" not in adata.uns:
        raise KeyError(
            "missing .uns['approximation_groupby'] information, this does not look like an approximation",
        )

    if isinstance(groupby, str):
        groupby = [groupby]

    groupby = list(groupby)
    if len(groupby) == 0:
        raise ValueError(
            "groupby must be a sequence with at least one element",
        )

    groupby_original = list(adata.uns["approximation_groupby"]["names"])
    for column in groupby:
        if column not in groupby_original:
            raise ValueError(
                f"Grouping not found in approximation: {column}",
            )

    # Trivial case
    if groupby_original == groupby:
        return adata.copy()

    indices_groupby = [groupby_original.index(gb) for gb in groupby]

    # Detect neighborhood
    neighborhood = "X_ncells" in adata.obsm
    if neighborhood:
        # We only have to touch the X_ncells, the rest is grouping agnostic
        multiindex = pd.Index(adata.uns["approximation_groupby"]["order"]).str.split(
            "\t", expand=True
        )
        multiindex.names = groupby_names
        ncells = pd.DataFrame(
            adata.obsm["X_ncells"],
            index=multiindex,
        )
        # FIXME: this probably needs more work
        ncells_cg = ncells.groupby(level=indices_groupby).sum()
        new_order = np.asarray(
            ["\t".join(x) for x in ncells_cg.index],
        )

        adata_cg = adata.copy()
        adata_cg.obsm["X_ncells"] = ncells_cs.values
        adata_cg.uns["approximation_groupby"] = {
            "names": groupby,
            "dtypes": [
                adata.uns["approximation_groupby"]["dtypes"][i] for i in indices_groupby
            ],
            "order": new_order,
        }

    else:
        # If not neighborhood, we have to actually change the matrix X and, if present,
        # the layers (e.g. avg and fractions)
        ncells_with_meta = adata.obs[list(groupby_original) + ["cell_count"]].copy()
        ncells_with_meta["idx"] = np.arange(len(ncells_with_meta))

        # To coarse grain, we need two things:
        # 1. the number of cells in each group
        # 2. the index of cells in each group
        gby = ncells_with_meta.groupby(groupby)

        X = np.zeros((gby.ngroups, adata.X.shape[1]), adata.X.dtype)
        obs_names = []
        if adata.layers:
            layers = {
                key: np.zeros((gby.ngroups, layer.shape[1]), layer.dtype)
                for key, layer in adata.layers.items()
            }

        for i, (groupid, group) in enumerate(gby):
            idx = group["idx"].values
            # Propagation involves multiplying expression by the number of cells in each subgroup, summing, dividing by the total
            # That can be recast as computing the fraction of the total and then collapsing the matrix against the frac vector
            fr_cell = 1.0 * group["cell_count"].values
            fr_cell /= float(fr_cell.sum())
            # Propagae the average
            X[i] = fr_cell @ adata.X[idx]
            if adata.layers:
                # Propagae the fraction_detected
                for key in layers:
                    layers[key][i] = fr_cell @ adata.layers[key][idx]
            # Stabilise the obs_names
            obs_name = "\t".join(str(gid) for gid in groupid)
            obs_names.append(obs_name)

        if adata.layers:
            adata_cg = anndata.AnnData(
                X=X,
                layers=layers,
                var=adata.var.copy(),
            )
        else:
            adata_cg = anndata.AnnData(
                X=X,
                var=adata.var.copy(),
            )

        # Obs names and metadata
        ncells_cg = gby.sum()["cell_count"]
        adata_cg.obs_names = obs_names
        adata_cg.obs["cell_count"] = ncells_cg.values
        for gb in groupby:
            adata_cg.obs[gb] = ncells_cg.index.get_level_values(gb)

        adata_cg.uns["approximation_groupby"] = {
            "names": groupby,
            "dtypes": [
                adata.uns["approximation_groupby"]["dtypes"][i] for i in indices_groupby
            ],
        }

    return adata_cg
